body {
    width: 600px;
    overflow-x: hidden;
}

.os-not-supported-info, .startup-failure-info {
    display: none;
}

body.os-not-supported #extensions-selection,
body.startup-failure #extensions-selection {
    display: none;
}

body.os-not-supported .os-not-supported-info {
    display: block;
}
body.startup-failure .startup-failure-info {
    display: block;
}

.not-supported {
    display: none !important;
}

.tabbar {
    display: block;
    padding: 0;
    /* color of separator line below heading */
    background-color: #d3d3d3;
    /* undo body side margins*/
    margin: 0 -17px 0px -17px;
    padding: 0px 17px;
}
.tabbar > li {
    display: inline-block;
    /* TODO draw a nice tab-like thingie */
    margin-top: 4px;
    border-top-left-radius: 4px;
    border-top-right-radius: 4px;
    border: 1px solid #999;
    border-bottom: none;
}
.tabbar > li > a {
    display: block;
    text-decoration: none;
    color: #333; /* can we just reset it to default text color?*/
    padding: 4px 8px;
}

.tabbar > li > a.active {
    font-weight: bold;
    background-color: #fff;
}
@media (prefers-color-scheme: dark) {
    .tabbar {
        background-color: #292c30;
    }
    .tabbar > li > a {
        color: #e3e3e3;
    }
    .tabbar > li > a.active {
        font-weight: bold;
        background-color: #111;
    }
}

.tab {
    display: none;
}

.tab.active {
    display: block;
}

.extensions-selection {
    padding: 0;
}
.extensions-selection > li {
    display: block;
    padding: 0;
}
.extensions-selection > li.dependent {
    padding-left: 20px;
}
.extensions-selection p {
    margin-top: 0;
    opacity: 0.7;
    line-height: 1.3;
    /* checkbox default width is 13px, try to align the description somewhat*/
    padding-left: 18px;
}
.extensions-selection textarea {
    width: 100%;
    white-space: nowrap;
    resize: vertical;
    min-height: 12lh;
}

.dialog-button-box {
    float: right;
    padding-bottom: 14px;
}


img.konqi {
    float: right;
    margin-right: -14px;
    width: 144px;
    height: 240px;
}
